# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""tests for the astng inference capabilities

Copyright (c) 2005-2006 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

__revision__ = "$Id: unittest_inference.py,v 1.12 2006-01-03 13:27:17 syt Exp $"

from os.path import join, abspath

from logilab.astng import builder, nodes, inference, utils
from logilab.common.testlib import TestCase, unittest_main

def get_name_node(start_from, name):
    return [n for n in start_from.nodes_of_class(nodes.Name) if n.name == name][0]

def get_node_of_class(start_from, klass):
    return start_from.nodes_of_class(klass).next()

builder = builder.ASTNGBuilder()

class InferenceUtilsTC(TestCase):

    def test_path_wrapper(self):
        infer_default = inference.path_wrapper(inference.infer_default)
        infer_end = inference.path_wrapper(inference.infer_end)
        self.failUnlessRaises(inference.InferenceError,
                              infer_default(1).next)
        self.failUnlessEqual(infer_end(1).next(), 1)
        
class InferenceTC(TestCase):

    DATA = '''
import exceptions

class C(object):
    "new style"
    attr = 4
    
    def meth1(self, arg1, optarg=0):
        var = object()
        print "yo", arg1, optarg
        self.iattr = "hop"
        return var
        
    def meth2(self):
        self.meth1(*self.meth3)
        
    def meth3(self, d=attr):
        b = self.attr
        c = self.iattr
        return b, c
    
ex = exceptions.Exception("msg")
v = C().meth1(1)
a, b, c = ex, 1, "bonjour"
[d, e, f] = [ex, 1.0, ("bonjour", v)]
g, h = f
i, (j, k) = u"glup", f

a, b= b, a # Gasp !
'''
        
    def setUp(self):
        self.astng = builder.string_build(self.DATA, __name__, __file__)

    def test_module_inference(self):
        infered = self.astng.infer()
        obj = infered.next()
        self.failUnlessEqual(obj.name, __name__)
        self.failUnlessEqual(obj.root().name, __name__)
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_class_inference(self):
        infered = self.astng['C'].infer()
        obj = infered.next()
        self.failUnlessEqual(obj.name, 'C')
        self.failUnlessEqual(obj.root().name, __name__)
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_function_inference(self):
        infered = self.astng['C']['meth1'].infer()
        obj = infered.next()
        self.failUnlessEqual(obj.name, 'meth1')
        self.failUnlessEqual(obj.root().name, __name__)
        self.failUnlessRaises(StopIteration, infered.next)

    def test_name_inference(self):
        infered = self.astng['C']['meth1']['var'].infer()
        var = infered.next()
        self.failUnlessEqual(var.name, 'object')
        self.failUnlessEqual(var.root().name, '__builtin__')
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_tupleassign_name_inference(self):
        infered = self.astng['a'].infer()
        exc = infered.next()
        self.failUnless(isinstance(exc, inference.Instance))
        self.failUnlessEqual(exc.name, 'Exception')
        self.failUnlessEqual(exc.root().name, 'exceptions')
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng['b'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, 1)
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng['c'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, "bonjour")
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_listassign_name_inference(self):
        infered = self.astng['d'].infer()
        exc = infered.next()
        self.failUnless(isinstance(exc, inference.Instance))
        self.failUnlessEqual(exc.name, 'Exception')
        self.failUnlessEqual(exc.root().name, 'exceptions')
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng['e'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, 1.0)
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng['f'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Tuple))
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_advanced_tupleassign_name_inference1(self):
        infered = self.astng['g'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, "bonjour")
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng['h'].infer()
        var = infered.next()
        self.failUnlessEqual(var.name, 'object')
        self.failUnlessEqual(var.root().name, '__builtin__')
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_advanced_tupleassign_name_inference2(self):
        infered = self.astng['i'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, u"glup")
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng['j'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, "bonjour")
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng['k'].infer()
        var = infered.next()
        self.failUnlessEqual(var.name, 'object')
        self.failUnlessEqual(var.root().name, '__builtin__')
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_swap_assign_inference(self):
        infered = self.astng.locals['a'][1].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, 1)
        self.failUnlessRaises(StopIteration, infered.next)
        infered = self.astng.locals['b'][1].infer()
        exc = infered.next()
        self.failUnless(isinstance(exc, inference.Instance))
        self.failUnlessEqual(exc.name, 'Exception')
        self.failUnlessEqual(exc.root().name, 'exceptions')
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_getattr_inference1(self):
        infered = self.astng['ex'].infer(path=[])
        exc = infered.next()
        self.failUnless(isinstance(exc, inference.Instance))
        self.failUnlessEqual(exc.name, 'Exception')
        self.failUnlessEqual(exc.root().name, 'exceptions')
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_getattr_inference2(self):
        infered = get_node_of_class(self.astng['C']['meth2'], nodes.Getattr).infer()
        meth1 = infered.next()
        self.failUnlessEqual(meth1.name, 'meth1')
        self.failUnlessEqual(meth1.root().name, __name__)
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_getattr_inference3(self):
        infered = self.astng['C']['meth3']['b'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, 4)
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_getattr_inference4(self):
        infered = self.astng['C']['meth3']['c'].infer()
        const = infered.next()
        self.failUnless(isinstance(const, nodes.Const))
        self.failUnlessEqual(const.value, "hop")
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_callfunc_inference(self):
        infered = self.astng['v'].infer()
        meth1 = infered.next()
        self.failUnless(isinstance(meth1, inference.Instance))
        self.failUnlessEqual(meth1.name, 'object')
        self.failUnlessEqual(meth1.root().name, '__builtin__')
        self.failUnlessRaises(StopIteration, infered.next)

    def test_args_default_inference1(self):
        optarg = get_name_node(self.astng['C']['meth1'], 'optarg')
        infered = optarg.infer()
        obj1 = infered.next()
        self.failUnless(isinstance(obj1, nodes.Const))
        self.failUnlessEqual(obj1.value, 0)
        self.failUnlessRaises(StopIteration, infered.next)

    def test_args_default_inference2(self):
        infered = self.astng['C']['meth3']['d'].infer(name='d')
        obj1 = infered.next()
        self.failUnless(isinstance(obj1, nodes.Const))
        self.failUnlessEqual(obj1.value, 4)
        self.failUnlessRaises(StopIteration, infered.next)
        
    def test_inference_restrictions(self):
        arg1 = get_name_node(self.astng['C']['meth1'], 'arg1')
        self.failUnlessRaises(inference.InferenceError, arg1.infer().next)

    def test_del(self):
        data = '''
del undefined_attr
        '''
        astng = builder.string_build(data, __name__, __file__)
        self.failUnlessRaises(inference.InferenceError,
                              astng.node.getChildNodes()[0].infer().next)
        
    def test_ancestors_inference(self):
        data = '''
class A:
    pass

class A(A):
    pass
        '''
        astng = builder.string_build(data, __name__, __file__)
        a1 = astng.locals['A'][0]
        a2 = astng.locals['A'][1]
        a2_ancestors = list(a2.ancestors())
        self.failUnlessEqual(len(a2_ancestors), 1)
        self.failUnless(a2_ancestors[0] is a1)

    def test_ancestors_inference2(self):
        data = '''
class A:
    pass

class B(A): pass

class A(B):
    pass
        '''
        astng = builder.string_build(data, __name__, __file__)
        a1 = astng.locals['A'][0]
        a2 = astng.locals['A'][1]
        a2_ancestors = list(a2.ancestors())
        self.failUnlessEqual(len(a2_ancestors), 2)
        self.failUnless(a2_ancestors[1] is a1)

    def test_lookup_limit(self):
        data = '''
l = [a
     for a,b in list]

a = 1
b = a
a = None

def func():
    c = 1
        '''
        astng = builder.string_build(data, __name__, __file__)
        names = astng.nodes_of_class(nodes.Name)
        a = names.next()
        stmts = a.lookup_name('a')[1]
        self.failUnlessEqual(len(stmts), 1)
        b = astng.locals['b'][1]
        self.failUnlessEqual(len(b.lookup_name('b')[1]), 2)
        self.failUnlessEqual(len(astng.lookup_name('b')[1]), 2)
        b_infer = b.infer()
        b_value = b_infer.next()
        self.failUnlessEqual(b_value.value, 1)
        self.failUnlessRaises(StopIteration, b_infer.next)
        func = astng.locals['func'][0]
        self.failUnlessEqual(len(func.lookup_name('c')[1]), 1)

    def test_f_arg_f(self):
        data = '''
def f(f=1):
    return f

a = f()
        '''
        astng = builder.string_build(data, __name__, __file__)
        a = astng['a']
        a_infer = a.infer()
        a_value = a_infer.next()
        self.failUnlessEqual(a_value.value, 1)
        self.failUnlessRaises(StopIteration, a_infer.next)
        
    def test_exc_ancestors(self):
        data = '''
def f():
    raise NotImplementedError
        '''
        astng = builder.string_build(data, __name__, __file__)
        names = astng.nodes_of_class(nodes.Name)
        nie = names.next().infer().next()
        self.failUnless(isinstance(nie, nodes.Class))
        nie_ancestors = [c.name for c in nie.ancestors()]
        self.failUnlessEqual(nie_ancestors, ['RuntimeError', 'StandardError', 'Exception'])

    def test_except_inference(self):
        data = '''
try:
    print hop
except NameError, ex:
    ex1 = ex
except Exception, ex:
    ex2 = ex
    raise
        '''
        astng = builder.string_build(data, __name__, __file__)
        ex1 = astng['ex1']
        ex1_infer = ex1.infer()
        ex1 = ex1_infer.next()
        self.failUnless(isinstance(ex1, inference.Instance))
        self.failUnlessEqual(ex1.name, 'NameError')
        self.failUnlessRaises(StopIteration, ex1_infer.next)
        ex2 = astng['ex2']
        ex2_infer = ex2.infer()
        ex2 = ex2_infer.next()
        self.failUnless(isinstance(ex2, inference.Instance))
        self.failUnlessEqual(ex2.name, 'Exception')
        self.failUnlessRaises(StopIteration, ex2_infer.next)

    def test_del(self):
        data = '''
a = 1
b = a
del a
c = a
a = 2
d = a
        '''
        astng = builder.string_build(data, __name__, __file__)
        n = astng['b']
        n_infer = n.infer()
        infered = n_infer.next()
        self.failUnless(isinstance(infered, nodes.Const))
        self.failUnlessEqual(infered.value, 1)
        self.failUnlessRaises(StopIteration, n_infer.next)
        n = astng['c']
        n_infer = n.infer()
        self.failUnlessRaises(inference.InferenceError, n_infer.next)
        n = astng['d']
        n_infer = n.infer()
        infered = n_infer.next()
        self.failUnless(isinstance(infered, nodes.Const))
        self.failUnlessEqual(infered.value, 2)
        self.failUnlessRaises(StopIteration, n_infer.next)

        

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()
    # unittest.main()
